#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#############################################################################
# File:          gitStatus.py - PyQt5 git repositories monitor
# Version:       2.0.2
# Author:        John Schwartzman, Forte Systems, Inc. 
# Last Revision: 06/23/2022
#############################################################################

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtGui import QColor
from PyQt5.QtCore import QDateTime, Qt, QEvent
from PyQt5.QtCore import QCommandLineParser, QCommandLineOption, QProcess
from PyQt5.QtWidgets import QListWidgetItem, QMenu, QDialog
from subprocess import Popen, PIPE
from configparser import ConfigParser
import sys, os

# string color constants
COLOR_BLACK  = '#000000'
COLOR_RED    = '#ff0000'
COLOR_GREEN  = '#00ff00'
COLOR_ORANGE = '#ffcc00'

INI_FILE_NAME = 'gitStatus.ini'

class Ui_dialog(QDialog):
    # member variables
    startDirs    = []
    exceptDirs   = []
    isVerbose    = False
    useTerminal  = False
    terminal     = ''
    standardView = ''
    imageView    = ''
    browserView  = ''
    
    def __init__(self):
        """
        This method initializes Ui_dialog's base class QDialog.
        """
        super().__init__()
        
    def setupUi(self, dialog):
        """
        This method was largely created by pyuic5 from GitStatusDialog.ui
        using the command 
        'pyuic5 -x -o gitStatusDialog.py GitStatusDialog.ui'.
        Seven lines were added manually before the last two lines 
        of this method.
        """
        dialog.setObjectName("dialog")
        dialog.setWindowModality(QtCore.Qt.NonModal)
        dialog.resize(700, 700)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap("git-gui.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        dialog.setWindowIcon(icon)
        dialog.setModal(False)
        self.gridLayout = QtWidgets.QGridLayout(dialog)
        self.gridLayout.setObjectName("gridLayout")
        self.repoView = QtWidgets.QListWidget(dialog)
        self.repoView.setObjectName("repoView")
        self.gridLayout.addWidget(self.repoView, 1, 0, 1, 1)
        self.dateTimeLabel = QtWidgets.QLabel(dialog)
        self.dateTimeLabel.setObjectName("dateTimeLabel")
        self.gridLayout.addWidget(self.dateTimeLabel, 2, 0, 1, 1)
        self.fileLabel = QtWidgets.QLabel(dialog)
        self.fileLabel.setObjectName("fileLabel")
        self.gridLayout.addWidget(self.fileLabel, 0, 1, 1, 3)
        self.closeButton = QtWidgets.QPushButton(dialog)
        self.closeButton.setObjectName("closeButton")
        self.gridLayout.addWidget(self.closeButton, 2, 3, 1, 1)
        self.refreshButton = QtWidgets.QPushButton(dialog)
        self.refreshButton.setObjectName("refreshButton")
        self.gridLayout.addWidget(self.refreshButton, 2, 2, 1, 1)
        self.repoLabel = QtWidgets.QLabel(dialog)
        self.repoLabel.setObjectName("repoLabel")
        self.gridLayout.addWidget(self.repoLabel, 0, 0, 1, 1)
        self.fileView = QtWidgets.QListWidget(dialog)
        self.fileView.setObjectName("fileView")
        self.gridLayout.addWidget(self.fileView, 1, 1, 1, 3)

        self.retranslateUi(dialog)
        
        # beginning of manually added lines
        self.fileView.installEventFilter(self)
        self.repoView.installEventFilter(self)
        self.repoView.itemSelectionChanged.connect(self.repoSelectionChanged)
        self.fileView.itemSelectionChanged.connect(self.fileSelectionChanged)
        self.refreshButton.clicked.connect(self.refresh)
        # end of manually added lines

        self.closeButton.clicked['bool'].connect(dialog.accept) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(dialog)

    def retranslateUi(self, dialog):
        """
        This method was created by pyuic5 from GitStatusDialog.ui
        """
        _translate = QtCore.QCoreApplication.translate
        dialog.setWindowTitle(_translate("dialog", "git Status Monitor"))
        self.repoLabel.setText(_translate("dialog", "repoLabel"))
        self.fileLabel.setText(_translate("dialog", "fileLabel"))
        self.dateTimeLabel.setText(_translate("dialog", "dateTImeLabel"))
        self.refreshButton.setText(_translate("dialog", "&Refresh"))
        self.closeButton.setText(_translate("dialog", "&Close"))
    
    def eventFilter(self, source, event):
        """
        handle context menu clicks (right click) 
          on items in repoView and fileView
        - create and display appropriate menu based on file type and status
        - create command
        - call runProcess(command, path) 
          Note: path may already be in the command. Use path = None (default)
        - call refresh() if source is self.repoView
        """
        contextMenu  = QMenu()
        if event.type() == QEvent.ContextMenu and source is self.fileView:
            # fileView context menu click
            path = self.repoView.currentItem().text()
            file = source.currentItem().text()
            ext  = file[1 + file.find('.'):]
            endOfFile = 2 + file.find(':')
            bDisplayMenu    = False
            imageAct        = None
            action          = None
            diffAct         = None
            normAct         = None
            browseAct       = None

            # add options to context menu
            if ext == 'png' or ext == 'jpg' or ext == 'bmp':
                # handle image file
                imageAct = contextMenu.addAction('image view')
            elif ext == 'html':
                # handle .html file
                browseAct = contextMenu.addAction('brower view')

            if file.startswith('modified: '):
                # handle modified files
                diffAct = contextMenu.addAction('difference view')
                endOfFile += 1
                bDisplayMenu = True
            elif file.startswith('untracked: ') or \
                 file.startswith('added: ') or     \
                 file.startswith('copied: ') or    \
                 file.startswith('renamed: '):
                # handle untracked, added, copied or renamed files
                endOfFile += 1
                bDisplayMenu = True
            elif endOfFile == 1:
                # handle other (normal) files
                bDisplayMenu = True
            
            if bDisplayMenu:
                # handle menu choice, determine and run process
                if not imageAct:
                    normAct = contextMenu.addAction('normal view')
                action = contextMenu.exec_(event.globalPos())
                if action != None:
                    if action == imageAct:  
                        # handle image files
                        command = self.imageView + ' ' + path + file[endOfFile:]
                        self.runProcess(command)
                        return True
                    elif action == browseAct:
                        # handle html files
                        command = self.browserView + ' ' + path + file[endOfFile:]
                        self.runProcess(command)
                        return True
                    elif action == normAct:
                        # handle normal files
                        command = self.standardView + ' ' + path + file[endOfFile:]
                        if self.useTerminal:
                            command = self.makeTerminalCommand(command)
                        self.runProcess(command)                                          
                        return True
                    elif action == diffAct:
                        # handle difference views of normal files
                        command = self.makeTerminalCommand('git diff  ' + \
                                                           file[endOfFile:])
                        self.runProcess(command, path)
                        return True
        elif event.type() == QEvent.ContextMenu and source is self.repoView:
            # repo view event click
            path = self.repoView.currentItem().text()
            action = None
            logAct = contextMenu.addAction('git log')
            terminalAct = contextMenu.addAction('terminal')
            action = contextMenu.exec_(event.globalPos())
            if action != None:
                if action == logAct:
                    self.runProcess(
                        self.makeTerminalCommand('git log'), path)
                elif action == terminalAct:
                    self.runProcess(
                        self.makeTerminalCommand(self.terminal), path)
                    self.refresh()
                return True
        return super().eventFilter(source, event)

    def hasGitDir(self, path):
        """
        This function returns True if path has a subdirectory
        named .git, otherwise it returns False.
        """
        return os.path.isdir(os.path.join(path, '.git'))

    def appendSlash(self, str):
        """
        This function returns 'str/'
        whether or not str ends with '/'
        """
        if not str.endswith('/'):
            str += '/'
        return str

    def makeTerminalCommand(self, command):
        """
        This method creates a string to run command in a terminal.
        """
        if self.terminal == 'gnome-terminal':
            # gnome-terminal needs to wait for child process to finish
            if command == self.terminal:
                command += ' --wait'
            else:
                command = self.terminal + " --wait -- bash -c \"" + \
                          command + "; exec bash \""
        else:   # self.terminal == 'konsole'
            if command != self.terminal:
                command = self.terminal + " -e bash -c \"" + command + ";"
        return command

    def runProcess(self, command, path = None):
        """
        - hide the dialog
        - create a process
        - set the process path (if necessary)
        - start process with command
        - show the dialog after process has finished
        """
        self.closeButton.parent().hide()    # hide the dialog
        process = QProcess()
        if path:
            process.setWorkingDirectory(path)
        process.start(command)
        process.waitForFinished(-1)
        process.close()
        self.closeButton.parent().show()    # show the dialog
        return

    def quickCheckPriority(self, dir):
        """
        use git status --porcelain to interrogate a repository
        returns COLOR_RED, COLOR_ORANGE or COLOR_GREEN (COLOR_BLACK if error)
        """
        p = Popen(["git","status","--porcelain"], cwd = dir, stdout = PIPE)
        p.wait()
        out = p.communicate()[0]
        if p.returncode != 0:
            return COLOR_BLACK          #error
        if out.find(b'M') != -1:
            return COLOR_RED            # modified - high priority
        if out.find(b'??') != -1:
            return COLOR_ORANGE         # untracked - medium priority
        if out.find(b'D') != -1:
            return COLOR_ORANGE         # deleted - medium priority
        if out.find(b'C') != -1:
            return COLOR_ORANGE         # copied - medium priority
        if out.find(b'A') != -1:
            return COLOR_ORANGE         # added - medium priority
        if out.find(b'R') != -1:        
            return COLOR_ORANGE         # renamed - medium priority
        return COLOR_GREEN              # no problems to report - low priority

    def refresh(self):
        """
        refreshButton has been clicked or refresh() was called
        - clear data structures and widgets
        - calls parseConfigFile()
        - calls populateDialog()
        - selects previously selected item in repoView if it still exists
        """
        # get text of currently selected item in repoView
        repoText = self.repoView.currentItem().text()
        self.startDirs.clear()
        self.exceptDirs.clear()
        self.repoView.clear()
        self.fileView.clear()
        if not self.parseConfigFile():    # read/create INI_FILE_NAME
            sys.exit(1)
        self.populateDialog()

        # find the text of the previously selected item
        # if the item exists, select it and scroll to it
        list = self.repoView.findItems(repoText, Qt.MatchExactly)
        if list:    # if list is empty this returns False
            self.repoView.setCurrentItem(list[0])
            self.repoView.scrollToItem(list[0])
        self.repoView.setFocus()

    def fileSelectionChanged(self):
        """
        fileView selection has changed
        - modify the fileView stylesheet to write border around selected file
        """
        self.fileView.setStyleSheet("""
        QListWidget::item:selected { color: """ + 
            self.fileView.currentItem().data(Qt.UserRole) + """; }
        QListWidget::item:selected { background-color: white; }
        QListWidget::item:selected { border: 2px solid red; }
        """)

    def repoSelectionChanged(self):
        """
        repoView selection has changed
        - modify the repoView stylesheet to write border arould selected repo
        - clear fileView
        - use git ls-files to get the list of files in repo
        - use git status --porcelain to check status of each file
        - populate fileView 
        """
        self.repoView.setStyleSheet("""
        QListWidget::item:selected { color: """ + 
            self.repoView.currentItem().data(Qt.UserRole) + """; }
        QListWidget::item:selected { background-color: white; }
        QListWidget::item:selected { border: 2px solid red; }
        """)
        self.fileView.clear()

        # get list of files in repository
        dir = self.repoView.currentItem().text()
        p = Popen(["git", "ls-files"], cwd = dir, stdout = PIPE)
        p.wait()
        out = p.communicate()[0]
        strArray = out.splitlines()

        i = 0
        sizeOfList = len(strArray)
        strList = []
        while i < sizeOfList:
            # leave space for status            
            s = strArray[i].rjust(3 + len(strArray[i])) 
            strList.append(s)
            i += 1

        # check for status of files
        p = Popen(["git", "status", "--porcelain"],
            cwd = dir, stdout = PIPE)
        p.wait
        out = p.communicate()[0]
        strArray = out.splitlines()

        # reconcile the 2 file lists
        i = j = 0
        itemsResolved = []
        # iterate over the list that contains status info
        while i <  len(strArray):
            found = False
            if itemsResolved.__contains__(i):
                i += 1
            while j < len(strList):
                a = strList[j][3:]
                b = strArray[i][3:]
                if a == b:  # file names match
                    # add the status information to strList
                    strList[j] = strArray[i]
                    itemsResolved.append(i)
                    found = True
                    break
                j += 1
            if not found:
                # might be untracked file
                strList.append(strArray[i])
            i += 1

        # color code the file list
        sizeOfList = len(strList)
        i = 0

        while i < sizeOfList:
            line = strList[i].decode()
            i += 1
            x = line[0] # char x indicaates changes ready for commit
            y = line[1] # char y indicates changes not yet staged for commit
            if x == 'M' or y == 'M':     # modified file
                s = "modified: "
                color = COLOR_RED
            elif x == 'D' or y == 'D':   # deleted file
                s = "deleted: "
                color = COLOR_ORANGE
            elif x == 'A' or y == 'A':   # added file
                s = "added: "
                color = COLOR_ORANGE
            elif x == 'R' or y == 'R':   # renamed file
                s = "renamed: "
                color = COLOR_ORANGE
            elif x == 'C' or y == 'C':   # copied file
                s = "copied: "
                color = COLOR_ORANGE
            elif x == '?' or y == '?':   # untracked file
                s = "untracked: "
                color = COLOR_ORANGE
            elif x == ' ' and y == ' ':  # normal file
                s = ""
                color = COLOR_GREEN

            # prepend status string
            item = QListWidgetItem(s + line[2:])
            item.setData(Qt.UserRole, color)
            item.setForeground(QColor(color))
            self.fileView.addItem(item)
            
        self.fileView.sortItems()
        # set the number of fileView items
        self.fileLabel.setText('Files: ' + str(self.fileView.count()))
        self.fileView.setCurrentRow(0)
    
    def populateDialog(self):
        """
        - display current date and time
        - walk the filesystem depth-first to search for git repositories
        - show added and ignored directories when in verbose mode
        - populate repoView with directories that have git repositories
        - alphabetize repoView
        - display number of repositories (nRepoCount)
        """
        dt = QDateTime.currentDateTime()
        self.dateTimeLabel.setText('Date/Time: ' +
                                   dt.toString(Qt.DefaultLocaleLongDate))
        nRepoCount = 0
        ignoredDirs = []
        # for each startDir walk the filesystem
        for start in self.startDirs:
            start = self.appendSlash(start)
            if not os.path.isdir(start):
                print('Error: ' + start + ' is not a directory.\n')
                sys.exit(1)

            if self.isVerbose:
                print('\nSearching for git repositories starting at ' 
                      + start + '...')

            # walk the filesystem to search for git repositories
            for root, dirs, __ in os.walk(start):
                if self.isVerbose:
                    for dir in dirs:
                        if dir in self.exceptDirs:
                            if not dir in ignoredDirs:
                                ignoredDirs.append(dir)
                                print('Ignoring: ' + dir)

                # prune dirs in place using list comprehension
                dirs[:] = [d for d in dirs if d not in self.exceptDirs]
                for dir in dirs:
                    thisDir = os.path.join(root, dir)

                    if thisDir.find('.git') != -1:    # we're inside a git dir
                        continue

                    if self.hasGitDir(thisDir):
                        # we found a directory that has a child .git directory 
                        # create a new QListWidgetItem
                        #  - set its text to thisDir
                        #  - set its data to the appropriate color for this item
                        #  - insert the item into repoView
                        item = QListWidgetItem(thisDir + '/')
                        item.setData(Qt.UserRole, self.quickCheckPriority(thisDir))
                        self.repoView.addItem(item)
                        nRepoCount += 1
                        if self.isVerbose:
                            print('Adding: ' + dir)
                        continue

        # after all items have been added to repoView,
        # 1. alphabetize the list
        self.repoView.setSortingEnabled(True)
        self.repoView.sortItems()

        # 2. set the foreground color for each item in the alphabetized list
        for i in range(nRepoCount):
            item = self.repoView.item(i)
            item.setForeground(QColor(item.data(Qt.UserRole)))

        self.repoView.setFocus();
        self.repoView.setCurrentRow(0)
        # set repo count
        self.repoLabel.setText('Git Repositories: ' + str(nRepoCount))
            
    def parseConfigFile(self):
        """
        - read ini file 
          or create ini file if it doesn't exist (provide reasonable values)
        - parse the ini file
        - load startDirs and exceptDirs
        - display startDirs and exceptDirs if in verbose mode
        - load and display terminal name
        - load and display standardView editor,
          imageView editor and browserView editor
        - return True if read or create worked properly
          or return False if read or create failed
        Note: failure will result from missing environment values or 
              missing executible files
        """
        parser = ConfigParser(allow_no_value=True)  # elements need not have values
        parser.optionxform = str    # ini file is case sensitive
        # we must have an absolute path to INI_FILE_NAME or ParseConfigFile will
        # create an ini file in the directory we're in when we run the program
        iniFile = os.path.abspath(os.path.join(os.path.dirname(sys.argv[0]),
                                               INI_FILE_NAME))
        try:
            with open(iniFile) as f:
                parser.read_file(f)
            
            if self.isVerbose:
                print('\nReading ' + iniFile + ' ...\n')

            for element in parser.sections():
                if element == 'startDirs':
                    for name, __ in parser.items(element):
                        if not name in self.startDirs:   # avoid duplicates
                            self.startDirs.append(name)
                if element == 'exceptDirs':
                    for name, __ in parser.items(element):
                        if not name in self.exceptDirs:  # avoid duplicates
                            self.exceptDirs.append(name)

            self.terminal     = parser['terminal']['name']
            self.standardView = parser['viewers']['standard']
            self.imageView    = parser['viewers']['image']
            self.browserView  = parser['viewers']['browser']

            if self.standardView == 'nano' or self.standardView == 'micro' or \
               self.standardView == 'vi' or self.standardView == 'vim':
                self.useTerminal = True # editor requires console

        except IOError: # ini file does not exist
            if self.isVerbose:
                print('\nWriting ' + iniFile + ' ...')
            parser.add_section('startDirs')
            envHome = os.getenv('HOME')
            if not envHome:
                print('\nError: environment variable HOME has not been set. '
                      'Please set it.\n')
                return False
            parser.set('startDirs', envHome)
            self.startDirs.append(envHome)
            parser.add_section('exceptDirs')

            parser.add_section('terminal')
            found = os.path.isfile('/usr/bin/konsole')  # does konsole exist?
            if found:
                parser.set('terminal', 'name', 'konsole')
                self.terminal = 'konsole'
            else:    # does gnome-terminal exist?
                found = os.path.isfile('/usr/bin/gnome-terminal')
                if found:
                    parser.set('terminal', 'name', 'gnome-terminal')
                    self.terminal = 'gnome-terminal' 
                else:
                    print('\nError: could not find terminal name.' +
                          'Please write it into gitStatus.ini.\n')
                    return False

            parser.add_section('viewers')
            envEditor = os.getenv('EDITOR')
            if not envEditor:
                print('\nError: Environment variable EDITOR has not been set. ' 
                      'Please set it.\n')
                return False
            self.standardView = envEditor[1 + envEditor.rfind('/'):] 
            parser.set('viewers', 'standard', self.standardView)
                        
            if self.standardView == 'nano' or self.standardView == 'micro' or \
               self.standardView == 'vi' or self.standardView == 'vim':
                self.useTerminal = True
  
            found = os.path.isfile('/usr/bin/okular') 
            if found:
                parser.set('viewers', 'image', 'okular')
                self.imageView = 'okular'
            else:
                found = os.path.isfile('/usr/bin/eog')
                if found:
                    parser.set('viewers', 'image', 'eog')
                    self.imageView = 'eog'
                else:
                    print('\nError: could not find image viewer name. '
                          'Please write it into gitStatus.ini.\n')
                    return False

            found = os.path.isfile('/usr/bin/firefox') or \
            	    os.path.isfile('/snap/bin/firefox')
            if found:
                parser.set('viewers', 'browser', 'firefox')
            else:
                found = os.path.isfile('/usr/bin/google-chrome')
                if found:
                    parser.set('viewers', 'browser', 'google-chrome')
                else:
                    print('\nError: could not find html browser. ' \
               	          'Please write it into gitStatus.ini.\n')
                    return False

            try:
                cfgfile = open(iniFile, 'w')
                parser.write(cfgfile)
                cfgfile.close()
            except IOError:
                print('\nError: Could not open ' + iniFile + 
                      ' for writing. Check file permissions.\n')
                return False

        self.terminal     = parser['terminal']['name']
        self.standardView = parser['viewers']['standard']
        self.imageView    = parser['viewers']['image']
        self.browserView  = parser['viewers']['browser']

        if self.isVerbose:
            # report on the contents of the ini file
            print('[startDirs]')
            for dir in self.startDirs:
                print(dir)
            print('[startDirs] contains ' + str(self.startDirs.__len__()) 
                + ' item(s).\n')
            print('[exceptDirs]')
            for dir in self.exceptDirs:
                print(dir)
            print('[exceptDirs] contains ' + str(self.exceptDirs.__len__()) 
                + ' item(s).\n')

            print('terminal name = ' + self.terminal)
            print('standardView  = ' + self.standardView)
            print('imageView     = ' + self.imageView)
            print('browserView   = ' + self.browserView + '\n')

        return True


if __name__ == "__main__":
    """
    main method
    - process command line options (--help, --version, --verbose)
    - create instance of application class (ui)
    - call setupUi to set dialog properties (calls retranslateUi)
    - call parseConfigFile
    - show the dialog
    - the application will process messages until it is told to exit
    """
    app = QtWidgets.QApplication(sys.argv)
    app.setApplicationName("gitStatus.py")
    app.setApplicationVersion("2.0.1")

    clp = QCommandLineParser()    # handle command line options
    clp.addHelpOption()
    clp.addVersionOption()
    verboseOption = QCommandLineOption("verbose", 
                                       "Issue progress messages to console.")
    clp.addOption(verboseOption)
    clp.process(sys.argv)

    dialog = QtWidgets.QDialog()    # create QDialog
    ui = Ui_dialog()                # instantiate ui
    ui.setupUi(dialog)              # configure all widgets in dialog
    ui.isVerbose = clp.isSet(verboseOption)

    if not ui.parseConfigFile():    # read or initialize/write ini file
        sys.exit(1)                 # exit if parseConfigFile() failed

    ui.populateDialog()             # populate the dialog's widgets

    dialog.show()                   # make the dialog visible
    sys.exit(app.exec_())           # handle all messages until exit
